# External
from typing import Union,List
# Internal Modules
from dbgen.support.datatypes.sqltypes     import Varchar,Date
from dbgen.support.datatypes.relation     import Relation
from dbgen.support.datatypes.insertupdate import Insert,InsertUpdate
from dbgen.support.datatypes.misc         import Arg,noIndex
from dbgen.support.datatypes.constraint   import NULL,IN
from dbgen.support.datatypes.rule         import Const,SimplePipe,PBlock,Rename,Unpack,SimpleFunc,Block

from dbgen.new_inputs.objects import *
from dbgen.core.lists         import flatten

from dbgen.scripts.IO.anytraj                   import anytraj
from dbgen.scripts.IO.parse_mendeleev           import parse_mendeleev
from dbgen.scripts.IO.get_job_metadata          import get_job_metadata
from dbgen.scripts.IO.populate_storage          import populate_storage
from dbgen.scripts.IO.get_catalog_logfiles      import get_catalog_logfiles

from dbgen.scripts.Pure.Misc.map_               import map_
from dbgen.scripts.Pure.Misc.hash_              import hash_
from dbgen.scripts.Pure.Misc.unzip              import unzip
from dbgen.scripts.Pure.Misc.triple             import triple
from dbgen.scripts.Pure.Misc.identity           import identity
from dbgen.scripts.Pure.Misc.concat_map         import concat_map
from dbgen.scripts.Pure.Misc.dftcode_dependent  import dftcode_dependent

from dbgen.scripts.Pure.Atoms.cell_info          import cell_info
from dbgen.scripts.Pure.Atoms.json_to_traj       import json_to_traj
from dbgen.scripts.Pure.Atoms.traj_to_json       import traj_to_json
from dbgen.scripts.Pure.Atoms.get_system_type    import get_system_type
from dbgen.scripts.Pure.Atoms.get_cell           import get_cell
from dbgen.scripts.Pure.Atoms.get_atom           import get_atom
from dbgen.scripts.Pure.Atoms.get_traj_gpaw_lazy import get_traj_gpaw_lazy
from dbgen.scripts.Pure.Atoms.get_traj_qe_lazy   import get_traj_qe_lazy
from dbgen.scripts.Pure.Atoms.get_traj_vasp_lazy import get_traj_vasp_lazy

from dbgen.scripts.Pure.Load.parse_pw_gpaw      import parse_pw_gpaw
from dbgen.scripts.Pure.Load.parse_pw_qe        import parse_pw_qe
from dbgen.scripts.Pure.Load.parse_pw_vasp      import parse_pw_vasp
from dbgen.scripts.Pure.Load.parse_xc_gpaw      import parse_xc_gpaw
from dbgen.scripts.Pure.Load.parse_xc_qe        import parse_xc_qe
from dbgen.scripts.Pure.Load.parse_xc_vasp      import parse_xc_vasp
from dbgen.scripts.Pure.Load.parse_psp_gpaw     import parse_psp_gpaw
from dbgen.scripts.Pure.Load.parse_psp_qe       import parse_psp_qe
from dbgen.scripts.Pure.Load.parse_psp_vasp     import parse_psp_vasp
from dbgen.scripts.Pure.Load.normalize_xc       import normalize_xc
from dbgen.scripts.Pure.Load.normalize_psp      import normalize_psp

Actions = List[Union[PBlock,List[PBlock]
              ,Block,List[Block]
              ,InsertUpdate,List[InsertUpdate]]]
Action = Union[PBlock,Block,InsertUpdate]

##############################################################################
# Constants
#----------
pop_job_cols = ['stordir','logfile','code','log','pwinp'
               ,'potcar','poscar','kptcar','paramdict']

mdata_input = ['stordir','code','paramdict']
mdata_cols  = ['user','timestamp','working_directory'
              ,'job_type','job_name']

elem_cols  =  [ 'element_id', 'symbol', 'name', 'atomic_weight','atomic_radius'
              , 'phase','evaporation_heat', 'pointgroup','spacegroup'
              , 'melting_point', 'metallic_radius', 'vdw_radius'
              , 'density', 'en_allen', 'is_radioactive'
              , 'lattice_struct' , 'fusion_heat'
              , 'econf', 'period', 'covalent_radius_bragg'
              , 'geochemical_class', 'abundance_crust', 'heat_of_formation'
              , 'electron_affinity', 'atomic_volume',  'boiling_point'
              , 'proton_affinity', 'covalent_radius_slater'
              , 'lattice_constant', 'dipole_polarizability'
              , 'en_ghosh', 'thermal_conductivity', 'group_id', 'en_pauling'
              , 'gas_basicity','abundance_sea']

cell_cols = ['a0','a1','a2','b0','b1','b2','c0','c1','c2']


dfts       = ['gpaw','qe','vasp']
parsed     = ['pw',  'xc','psp']

calc_funcs = [parse_pw_gpaw,parse_pw_qe,parse_pw_vasp # type: ignore
             ,parse_xc_gpaw,parse_xc_qe,parse_xc_vasp
             ,parse_psp_gpaw,parse_psp_qe,parse_psp_vasp
             ,normalize_xc,normalize_psp,identity]

traj_funcs = [get_traj_gpaw_lazy,get_traj_qe_lazy,get_traj_vasp_lazy,identity
             ,unzip,traj_to_json,get_atom,get_cell,hash_]

##############################################################################
# Complicated Functions
#----------------------

relax_calc_acts = [PBlock(func = dftcode_dependent
                          ,name = x
                          ,args = noIndex(['dftcode'
                                          ,'log'
                                          ,'parse_%s_gpaw'%x
                                          ,'parse_%s_qe'%x
                                          ,'parse_%s_vasp'%x
                                          ,'identity'
                                          ,'identity' if x =='pw' else 'normalize_%s'%x]))
                    for x in parsed] + [PBlock(identity,'dftcode',[Arg('code')])]



traj_acts = [PBlock(triple
                   ,'anytraj_log_poscar'
                   ,args = noIndex(['anytraj','log','poscar']))

            ,PBlock(dftcode_dependent
                   ,name = 'get_traj'
                   ,args = noIndex(['code'
                                   ,'anytraj_log_poscar'
                                   ,'get_traj_gpaw_lazy'
                                   ,'get_traj_qe_lazy'
                                   ,'get_traj_vasp_lazy'
                                   ,'identity'
                                   ,'unzip'])) # returns ([ASEATOMS]m,[STEPNUM]m,[ENERGY]m,[TRAJATOMLIST]m)
            ,PBlock(map_
                    ,name = 'cell'
                    ,args = noIndex(['get_cell','structs_few'])) # returns ([ax]m,[ay]m,...)

           ,PBlock(map_
                    ,name = 'structs_few'
                    ,args = [Arg('traj_to_json')
                            ,Arg('get_traj', 0)]) # returns [raw_json]m
           ,PBlock(map_
                    ,name = 'rawhash'
                    ,args = [Arg('hash_')
                            ,Arg('structs_few')]) # returns [hashed raw_json]m

           ,PBlock(concat_map
                  ,name = 'atoms'
                  ,args = noIndex(['get_atom','structs_few'])) # returns ([ASEATOMS]n,[Ind]n,[ELEMENT]n,[X]n,[Y]n,...)

           ,PBlock(identity
                ,name = 'structs_many'
                ,args = [Arg('atoms', 0)])# returns [raw_json]n


           ,PBlock(flatten
                 ,name = 'flattened_trajatoms'
                 ,args = [Arg('get_traj',3)])  # [STEPNUM,Ind,FX,FY,FZ]n

           ,PBlock(unzip
                     ,'trajatoms'
                     ,args=[Arg('flattened_trajatoms')])

           ,Rename('structs_few','raw')     # ([STEPNUM]n,[Ind]n,[FX]n,[FY]n,[FZ]n)
           ,Unpack('cell',cell_cols)
           ,Cell.insert(cell_cols)
           ,Struct.insert(['raw','rawhash'])
           ,Struct.update_component(Cell.select(cell_cols))] # type: Actions

########################################################################
# Relations
#----------
loading_relations = [
    Relation(name    = 'catalog'
            ,view    = view(attr  = [Roots.root]                    # type: ignore
                           ,const = [Roots.label  == "'catalog'"    # type: ignore
                                    ,Roots.active == 1])      # type: ignore
            ,actions = [SimpleFunc(func    = get_catalog_logfiles
                                 ,inputs  = ['root']
                                 ,outputs = pop_job_cols)
                        ,Job.insert(pop_job_cols)])

   ,Relation(name    = 'populate_storage'
            ,actions = [SimpleFunc(func    = populate_storage
                                  ,outputs = ['root','code','label'])
                       ,Roots.insert(['root','code','label'])])

   ,Relation(name    = 'job_metadata'
            ,view    = view(attr  = Job.select(mdata_input)         # type: ignore
                           ,const = [NULL(Job.user)])               # type: ignore
            ,actions = [SimpleFunc(func    = get_job_metadata
                                 ,inputs  = mdata_input
                                 ,outputs = mdata_cols)
                       ,Job.update(mdata_cols)])

   ,Relation(name   = 'anytraj'
            ,view   = view(attr  = Job.select(['stordir']) # type: ignore
                          ,const = [NULL(Job.anytraj)])  # type: ignore
            ,actions   = [SimpleFunc(anytraj,['stordir'],['anytraj'])
                         ,Job.update(['anytraj'])])

    ,Relation(name = 'mendeleev'
             ,actions = [SimpleFunc(parse_mendeleev,outputs = elem_cols)
                        ,Element.insert(elem_cols)])

    ,Relation(name    = 'cell_info'
             ,view    = view(attr =Cell.select(cell_cols)) # type: ignore
             ,actions = [SimpleFunc(cell_info,cell_cols,['surface_area','volume'])
                        ,Cell.update(['surface_area','volume'])])

   ,Relation(name    = 'relax_calc'
            ,view    = view(attr =Job.select(['log','code'])         # type: ignore
                           ,const = [Relax_job._hasnt(Calc)])      # type: ignore
            ,consts  = {f.__name__:Const(f) for f in calc_funcs}
            ,actions = [Relax_job.update_component(
                            Calc.select(['dftcode','pw','xc','psp']))])

   ,Relation(name = 'systype'
            ,view =  view(attr  = [Struct.raw]                 # type: ignore
                         ,const = [NULL(Struct.system_type)])  # type: ignore
            ,actions = [SimplePipe([json_to_traj,get_system_type]
                               ,inputs  = ['raw']
                               ,outputs = ['system_type'])
                        ,Struct.update(['system_type'])])

   ,Relation(name    = 'relax_job'
            ,view  = view(attr  = Job.select()        # type: ignore
                         ,const = [IN(Job.job_type    # type: ignore
                                    ,["'relax'","'latticeopt'","'vcrelax'"])])
            ,actions = [Relax_job.insert()])

    ,Relation(name    = 'traj'
             ,view    = view(Job.select(['code','log','anytraj','poscar','stordir'])) # type: ignore
             ,consts  = {f.__name__:Const(f) for f in traj_funcs}
             ,actions = traj_acts )
#     ,metatemp = MetaTemp([
#         Block(name = 'insert_cell'
#              ,text = mkInsCmd('cell',cell_cols,sqlite=False)
#              ,args = cell_args)
#
#        ,Block(name = 'query_cell'
#              ,text = ('SELECT C.cell_id FROM cell AS C WHERE '
#                      +addQs(['C.%s'%x for x in cell_cols],' AND '))
#              ,args = cell_args)
#
#        ,Block(name = 'insert_struct'
#              ,text = """INSERT INTO struct (raw,rawhash,cell_id)
#                            VALUES (%s,SHA2(%s,512),%s)
#                         ON DUPLICATE KEY UPDATE cell_id=cell_id"""
#              ,args = [Arg('structs_few')
#                       ,Arg('structs_few')
#                      ,Arg('query_cell',0)])
#
#        ,Block(name = 'query_struct_few'
#              ,text = 'SELECT S.struct_id FROM struct AS S WHERE S.rawhash = SHA2(%s,512)'
#              ,args = [Arg('structs_few')])
#
#        ,Block(name = 'query_struct_many'
#              ,text = 'SELECT S.struct_id FROM struct AS S WHERE S.rawhash = SHA2(%s,512)'
#              ,args = [Arg('structs_many')])
#
#        ,Block(name = 'insert_atom'
#              ,text = mkInsCmd('atom',['struct_id','atom_id','element_id'
#                                   ,'x','y','z','constrained','magmom','tag']
#                             ,sqlite=False)
#              ,args = [Arg('query_struct_many')]
#                     +[Arg('atoms',i) for i in range(1,9)])
#
#      ,Block(name = 'insert_traj'
#            ,text = mkInsCmd('traj',['job_id','traj_id','struct_id','energy','final'],sqlite=False)
#            ,args = [Arg('job_id')
#                    ,Arg('get_traj',1)
#                    ,Arg('query_struct_few')
#                    ,Arg('get_traj',2)
#                    ,Arg('get_traj',4)])
#
#     ,Block(name = 'insert_trajatom'
#           ,text = mkInsCmd('trajatom',['job_id','traj_id','struct_id','atom_id','fx','fy','fz'],sqlite=False)
#           ,args = [Arg('job_id')
#                   ,Arg('trajatoms',0)
#                   ,Arg('query_struct_many')]
#                  +[Arg('trajatoms',i) for i in range(1,5)])]))


]


relations = loading_relations



#External
from typing import Dict,Any,Tuple,Union,Set,List,Optional
from networkx    import DiGraph                                 # type: ignore
from networkx.algorithms.dag          import descendants        # type: ignore
from networkx.algorithms.simple_paths import all_simple_paths   # type: ignore

# Internal
from dbgen.support.datatypes.sqltypes     import SQLType
from dbgen.support.datatypes.attr         import Attr
from dbgen.support.datatypes.table        import Table,FK as OldFK
from dbgen.support.datatypes.constraint   import EQ,NE,LT,GT,NULL,Constraint
from dbgen.support.datatypes.insertupdate import Insert,Update,InsertFK,InsertUpdate
from dbgen.support.datatypes.rule         import Block,Rename
from dbgen.scripts.Pure.Misc.identity     import identity
from dbgen.support.datatypes.misc         import Arg

from dbgen.core.lists import flatten

#############################################################################
# Constants
#----------
class NotNull(object): pass
NOTNULL = NotNull()
class Unique(object): pass
UNIQUE = Unique()

class DEFAULT(object):
    def __init__(self,val:Any)->None:
        self.val = val
################################################################################
# Main Class
#------------


class Object(object):
    """
    Superclass for user-defined objects (which correspond to DB tables)

    _parents
        -- SUFFICIENT and NECCESARY objects needed to specify an instance
          (e.g. for 'Atom', 'Struct' and 'Element' (+ 'atom_id')
           are both necessary, but Struct+atom_id are sufficient.
    _components
        --

    <anything else>
        -- normal attribute (doesn't start with '_')
        -- should provide a SQLType followed by other information, if necessary

    """
    universe = {}    # type: Dict[str,Object]
    todo     = set() # type: Set[str]
    _parents = []    # type: List[Object]
    _name    = ''
    _pks     = []    # type: List[Attr]
    deleted  = Attr('deleted')

    def __init__(self,alias:Optional[str]=None)->None:
        if alias is None:
            alias = self._name # type: ignore
        self._alias=alias # only defining property of an INSTANCE of a class

    @classmethod
    def _process_parent(cls,par:type)->None:
        new_pks = [Attr(par._name+'_'+k.name,pk=True,obj=cls._name)  # type: ignore
                        for k in par._pks]         # type: ignore
        cls._pks.extend(new_pks)                   # type: ignore
        cls._fks.append(FK(new_pks,par,par._pks))  # type: ignore
        for p in new_pks: cls._addAttr(p)
    @classmethod
    def _process_component(cls,comp:type)->None:
        new_fks = [Attr(comp._name+'_'+k.name,obj=cls._name)  # type: ignore
                        for k in comp._pks]         # type: ignore
        for f in new_fks: cls._addAttr(f)
        cls._fks.append(FK(new_fks,comp,comp._pks))  # type: ignore

    @classmethod
    def _init_vars(cls)->None:
        """
        Initialize the parents,components,many_to_one,desc,pks, and fks fields
        Also take all attributes and make them fields of the class
        """
        import pdb;pdb.set_trace()
        cd               = cls.__dict__
        name             = cls.__name__.lower()  # type: ignore
        cls._parents     = cd.get('_parents',[])                     # type: ignore
        cls._components  = cd.get('_components',[])                  # type: ignore
        cls._many_to_one = cd.get('_many_to_one',False)              # type: ignore
        cls._desc        = cls.__dict__.get('__doc__','(No description)') # type: ignore

        cls._pks,cls._fks,cls._attrs = [],[],{}  # type: ignore

        new_attrs  = [cls._mkAttr(k,v)       # type: ignore
                      for k,v in cd.items()
                      if k[0]!='_'] + [Attr('deleted',obj=name,default=0)]

        has_indep_pk = (not cls._parents) or cls._many_to_one # type: ignore
        if has_indep_pk:
            pk = Attr(name+'_id',pk=True,auto=True,obj=name)
            cls._pks = [pk]          # type: ignore
            new_attrs.append(pk)

        for a in new_attrs:
            cls._addAttr(a)

    @classmethod
    def _addAttr(cls,attr:Attr)->None:
        if attr.name in cls._attrs:         # type: ignore
            print('Cannot re-add '+str(attr))
        else:
            assert(attr.obj == cls._name)  # type: ignore
            cls._attrs[attr.name] = attr   # type: ignore
            setattr(cls,attr.name,attr)

    @classmethod
    def _mkAttr(cls
               ,name : str
               ,data : Union[SQLType,Tuple]
               ) -> Attr:
        """
        Take user input and construct and Attr
        """
        if isinstance(data,tuple):
            types = (SQLType,NotNull,DEFAULT,Unique)
            assert all([isinstance(x,types) for x in data])
            dtype = [d for d in data if isinstance(d,SQLType)][0]
            default = [d.val for d in data if isinstance(d,DEFAULT)]
            return Attr(name
                       ,dtype   = dtype
                       ,nnull   = NOTNULL in data
                       ,uniq    = UNIQUE in data
                       ,obj     = cls._name                         # type: ignore
                       ,default = default[0] if default else None)
        else:
            err = 'Expected field %s to contain dtype, but instead found %s %s'
            assert isinstance(data,SQLType),err%(name,data,type(data))
            return Attr(name,dtype=data,obj=cls._name) # type: ignore

    @classmethod
    def _mkView(cls
               ,attr  : List[Union['Object',Attr]]
               ,const : List[Constraint]            = []
               ,nn    : List['Object']              = []
               ) -> 'View':
        return View(cls.universe,attr,const,nn)

    @classmethod
    def _get_unsafe(cls,colname:str)->Attr:
        return cls._attrs[colname] # type: ignore

    @classmethod
    def insert(cls,cols:Optional[List[str]] = None)->InsertUpdate:
        if cols is None:
            return Insert(cls.universe,cls._pks) # type: ignore
        else:
            try:
                return Insert(cls.universe,[cls._get_unsafe(x) for x in cols])
            except ValueError as e:
                print(e)
                import pdb;pdb.set_trace()
                raise ValueError
    @classmethod
    def select(cls
              ,attrs  : Optional[List[str]] = None
              ) -> List[Attr]:
        if attrs is None:
            return cls._pks # type: ignore
        else:
            return list(map(cls._get_unsafe,attrs))

    @classmethod
    def update(cls,attrs : List[str])->Update:
        return Update(cls.universe,[cls._get_unsafe(x) for x in attrs])

    @classmethod
    def update_component(cls
                        ,attrs   : List[Attr]
                        ) -> InsertUpdate:
        """
        Need to control the INSERT, get PK, update PK process
        """
        return InsertFK(cls.universe,attrs,cls)

    @classmethod
    def _mkTable(cls)->Table:
        return Table(name = cls._name,desc = cls._desc                       # type: ignore
                    ,cols = [c._mkCol() for c in cls._attrs.values()]        # type: ignore
                    ,fks  = [fk.to_old_style() for fk in cls._fks])          # type: ignore
    @classmethod
    def _hasnt(cls,comp:type)->Constraint:
        for fk in cls._fks: # type: ignore
            if fk._to == comp:
                return NULL(fk.from_attr[0])
        raise ValueError

def new_model(name:str)->Object:
    """
    Create a new model
    """
    class New(Object):
        _modelname = name

        def __init_subclass__(cls      : type
                             ,**kwargs : dict
                             ) -> None:
            """
            Upgrade a user-defined class given the state of the universe
            """
            if kwargs:
                raise ValueError('kwargs? '+str(kwargs))
            else:
                cls._init_subclass() # type: ignore

        @classmethod
        def _init_subclass(cls) -> None:
            uni  = cls.universe
            todo = cls.todo
            name = cls.__name__.lower()

            cls._name = name; cls._alias = name  # type: ignore
            if name in uni:
                print('Warning: this class (%s) was already defined for this universe'%name)
                return None # No repeats allowed

            cls._init_vars() # type: ignore


            for p in cls._parents: # type: ignore
                assert isinstance(p,(type,str)), type(p) # basic integrity check

                # Handle this if it's a real definition or known string
                #------------------------------------------------------------
                if isinstance(p,type) or p.lower() in uni:
                    if isinstance(p,type):
                        parent = p
                    else:
                        parent = uni[p.lower()]

                    # Acknowledge only after all of its parents have been provided
                    #------------------------------------------------------------
                    if all([isinstance(pp,type) for pp in parent._parents]): # type: ignore
                        cls._process_parent(parent)                      # type: ignore
                else:
                    todo.add(p.lower())

            for c in cls._components: # type: ignore
                assert isinstance(c,(type,str)), type(c) # basic integrity check

                # Handle this if it's a real definition rather or known string
                #------------------------------------------------------------
                if isinstance(c,type) or c.lower() in uni:
                    if isinstance(c,type):
                        comp = c
                    else:
                        parent = uni[c.lower()]

                    # Acknowledge only after all of its parents have been provided
                    #------------------------------------------------------------
                    if all([isinstance(pp,type) for pp in cls._parents]):
                        cls._process_component(comp)
                else:
                    assert isinstance(p,str)
                    todo.add(p.lower())

            # Wrap things up after we're done
            #--------------------------------
            uni[name] = cls # type: ignore
            if name in todo:
                # UPGRADE EVERYTHING ELSE IN UNIVERSE THAT POINTS TO THIS VIA STR
                todo.remove(name)

    return New # type: ignore



#mike = new_model('mike') # type: Any

################################################################################
# HELPER CLASSes
#---------------
class FK(object):
    """
    A Foreign Key in the language of Objects/Attributes is a pair of lists of
    Attributes, each having equal length but referring to different Objects
    """
    def __init__(self
                ,from_attr : List['Attr']
                ,_to       : Object
                ,to_attr   : List['Attr']
                ) -> None:
        # Validate
        #---------
        err1 = 'FK args not of equal length: %s %s'
        assert len(from_attr) == len(to_attr), err1%(from_attr,to_attr)
        err2 = 'FK args share the same object: %s'
        assert from_attr[0].obj != to_attr[0].obj, err2%from_attr[0].obj

        # Store Fields
        #-------------
        self.from_attr = from_attr
        self._to       = _to
        self.to_attr   = to_attr

    def __str__(self)->str:
        return self.zipped()
    def __repr__(self)->str:
        return str(self)

    def zipped(self) -> str:
        z = zip(self.from_attr,self.to_attr)
        return ' AND '.join(['%s = %s'%(f,t) for f,t in z])

    def to_old_style(self) -> OldFK:
        if self.to_attr[0].obj is None:
            raise ValueError
        else:
            return OldFK([a.name for a in self.from_attr]
                         ,self._to._name # type: ignore
                         ,[a.name for a in self.to_attr])

    def rename(self)->List[Block]:
        return [Rename(y.name,x.name)
                    for x,y in zip(self.from_attr,self.to_attr)]

################################################################################

class Join(object):
    """
    JOIN <A> AS <B> ON <C> = <D> AND ...
    """
    def __init__(self
                ,obj  : Object
                ,fk   : FK
                ,kind : str     = 'INNER'
                ) -> None:
        self.obj  = obj
        self.fk   = fk
        self.kind = kind
        self.direction = True # flip this to change direction of join

    # Basic Methods
    #--------------
    def __str__(self) -> str:
        join   = '%s JOIN'%self.kind
        to_obj = self.fk._to if self.direction else self.obj
        if to_obj is None:
            raise ValueError
        else:
            tab    = '%s AS %s'%(to_obj._name,to_obj._alias)
            return '\n\t%s %s \n\t\tON %s'%(join,tab,self.fk.zipped())

    def flip(self)->None:
        self.direction = not self.direction



class From(object):
    """
    FROM clause
    """
    def __init__(self
                ,obj   : Object
                ,joins : List[Join] = []
                ) -> None:
        self.obj   = obj
        self.joins = joins
        self.allobj = [obj]+[j.obj for j in joins]
        assert len(set(self.allobj)) == len(joins)+1

    # Basic Methods
    #--------------
    def __str__(self) -> str:
        tab = ' %s AS %s '%(self.obj._name,self.obj._alias) # type: ignore
        return 'FROM ' + tab + ''.join(['\n%s'%j for j in self.joins])

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self,other : object)->bool:
        if isinstance(other,Object):
            return self.__dict__ == other.__dict__
        else:
            return False



##############################################################################
class View(object):
    """
    A query of sorts
    """
    def __init__(self
                ,uni         : Dict[str,Object]
                ,attrs       : List[Union[Attr,Object]]
                ,constraints : List[Constraint]         = []
                ,mbNull      : List[Object]               = []
                ) -> None:

        self.uni     = uni
        self.objects = list(uni.values())
        self.graph   = self.mk_graph()

        allobj = sorted(set([uni[x.obj] if isinstance(x,Attr) else x
                             for x in attrs]))
        self.allobj = allobj

        pk_attrs = flatten([o._pks for o in allobj])

        self.attrs = pk_attrs + [x for x in attrs if isinstance(x,Attr)]
        self.constraints = constraints + [o.deleted==0 for o in allobj]
        c_attrs = flatten([list(c.attrs()) for c in constraints])
        all_attr = self.attrs + c_attrs

        self._from = self.mkFrom(all_attr)

        for j in self._from.joins:
            if j.fk._to in mbNull:
                j.kind = "LEFT"


    def __str__(self)->str:
        cols   = '\n\t,'.join([str(a) for a in self.attrs])
        consts = '\n\tAND '.join([str(c) for c in self.constraints])

        return 'SELECT %s\n%s\nWHERE %s'%(cols,self._from,consts or '1')


    def mkFrom(self,attrs:List[Attr])->From:
        """
        Construct a FROM clause that joins the right tables in order to get
        access to a set of attributes mentioned by some query
        """
        objects  = set([a.obj for a in attrs])

        found    = False
        for obj in objects:
            desc = descendants(self.graph,obj)
            if all([o in desc for o in objects if o is not obj]):
                found = True
                first = obj

        if not found:
            import pdb;pdb.set_trace()
            raise ValueError('You screwed up')

        joins = []
        for o in objects:
            if o != first:
                paths = self.path(first,o)
                lp    = len(paths)

                if lp != 1:
                    print("Warning, %d paths found between %s and %s"%(lp,first,o))
                path = paths[0]
                assert path[0] == first
                curr_obj = self.uni[first] # initialize with this

                for s in path[1:]: # first element = first._name
                    next_obj = self.uni[s]
                    fk = self.get_fk(curr_obj,next_obj)
                    joins.append(Join(next_obj,fk))
                    curr_obj = next_obj

        return From(self.uni[first],joins)

    def mk_graph(self)->DiGraph:
        """
        Given parent/component relationships between objects, create a Directed
        Graph that captures the possible flows of information.
        """
        G = DiGraph()
        G.add_nodes_from([o._name for o in self.objects])    # type: ignore
        for o1 in self.objects:        # type: ignore
            if not o1._many_to_one:     # type: ignore
                for o2 in o1._parents:   # type: ignore
                    G.add_edge(o2._name,o1._name)   # type: ignore
            for o2 in o1._parents + o1._components:   # type: ignore
                    G.add_edge(o1._name,o2._name)    # type: ignore
        return G

    def path(self,o:str,o2:str)->List[List[str]]:
        """
        All possible paths between two objects via FK propagation
        """
        paths = all_simple_paths(self.graph,o,o2)
        return sorted(list(paths),key=len)

    def get_fk(self,o1:Object,o2:Object)->FK:
        """
        Gets a FK between two Objects -- error if more/fewer are found
        """
        fks = [fk for fk in o1._fks if fk._to is o2]  + [fk for fk in o2._fks if fk._to is o1]    # type: ignore
        if len(fks) != 1:
            raise ValueError('%d many fks between %s and %s: %s'%(len(fks),o1,o2,fks))
        else:
            return fks[0]
